@@ -14,7 +14,7 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-/* Copyright (C) 2004-2024 Erik Auerswald <auerswal@unix-ag.uni-kl.de> */
+/* Copyright (C) 2004-2025 Erik Auerswald <auerswal@unix-ag.uni-kl.de> */
 /* Copyright (C) 2013 Cristiano Fontana <fontanacl@ornl.gov> */
 
 /* ImLib2 Header */
@@ -32,7 +32,7 @@
 
 /* option parsing */
 #include <getopt.h>         /* getopt */
-#include <unistd.h>         /* getopt */
+#include <unistd.h>         /* getopt, read, write, STDIN_FILENO */
 
 /* file permissions */
 #include <sys/stat.h>       /* umask */
@@ -57,8 +57,8 @@ static char * tmp_imgfile(unsigned int flags)
   char *name;
   size_t pattern_len;
   int handle;
-  unsigned char buf;
-  ssize_t count = 0;
+  unsigned char buf[BUFSIZ];
+  ssize_t read_count = 0, write_count = 0;
   size_t pat_suffix_len = strlen(DIR_SEP TMP_FILE_PATTERN);
   size_t dir_len;
 
@@ -99,12 +99,16 @@ static char * tmp_imgfile(unsigned int flags)
   }
 
   /* copy image data from stdin to tmp file */
-  while((fread(&buf, sizeof(char), 1, stdin)) > 0) {
-    count = write(handle, &buf, 1);
-    if (count <= 0) break;
+  while((read_count = read(STDIN_FILENO, buf, sizeof(buf))) > 0) {
+    write_count = write(handle, buf, read_count);
+    if (write_count <= 0) break;
   }
   close(handle); /* filehandle is no longer needed, Imlib2 uses filename */
-  if(ferror(stdin) || (count <= 0)) {
+  if(read_count < 0) {
+    perror(PROG ": could not read image from standard input");
+    exit(99);
+  }
+  if(write_count <= 0) {
     perror(PROG ": could not copy image data to temporary file");
     unlink(name);
     exit(99);
@@ -114,16 +118,20 @@ static char * tmp_imgfile(unsigned int flags)
 }
 
 /* return number of foreground pixels in a scanline */
-static unsigned int scanline(Imlib_Image *image, Imlib_Image *debug_image,
+static unsigned int scanline(Imlib_Image *debug_image,
                              int x, int y, int len, direction_t dir,
                              color_struct d_color, double thresh,
                              luminance_t lt, unsigned int flags)
 {
-  Imlib_Color imlib_color;
+  Imlib_Color imlib_color, debug_color;
   int lum, i, ix=x, iy=y, start, end;
   unsigned int found_pixels = 0;
   start = (dir == HORIZONTAL) ? x : y;
   end = start + len;
+  debug_color.red = d_color.R;
+  debug_color.green = d_color.G;
+  debug_color.blue = d_color.B;
+  debug_color.alpha = d_color.A;
   for (i = start; i <= end; i++) {
     if (dir == HORIZONTAL) ix = i;
     else iy = i;
@@ -131,10 +139,7 @@ static unsigned int scanline(Imlib_Image *image, Imlib_Image *debug_image,
     lum = get_lum(&imlib_color, lt);
     if(is_pixel_set(lum, thresh)) {
       if(flags & USE_DEBUG_IMAGE) {
-        imlib_context_set_image(*debug_image);
-        imlib_context_set_color(d_color.R, d_color.G, d_color.B, d_color.A);
-        imlib_image_draw_pixel(ix, iy, 0);
-        imlib_context_set_image(*image);
+        draw_color_pixel(debug_image, ix, iy, debug_color);
       }
       found_pixels++;
     }
@@ -334,10 +339,11 @@ int main(int argc, char **argv)
       {"print-spaces", 0, 0, 's'}, /* print spaces between distant digits */
       {"space-factor", 1, 0, 'A'}, /* relative distance to add spaces */
       {"space-average", 0, 0, 'G'}, /* avg instead of min dst for spaces */
+      {"adapt-after-crop", 0, 0, 'F'}, /* don't adapt threshold before crop */
       {0, 0, 0, 0} /* terminate long options */
     };
     c = getopt_long (argc, argv,
-                     "hVt:vaTn:N:i:d:r:m:M:o:O:D::pPf:b:Igl:SXCc:H:W:sA:G",
+                     "hVt:vaTn:N:i:d:r:m:M:o:O:D::pPf:b:Igl:SXCc:H:W:sA:GF",
                      long_options, &option_index);
     if (c == -1) break; /* leaves while (1) loop */
     switch (c) {
@@ -618,6 +624,13 @@ int main(int argc, char **argv)
                           flags & SPC_USE_AVG_DST);
         }
         break;
+      case 'F':
+        flags |= ADAPT_AFTER_CROP;
+        if(flags & DEBUG_OUTPUT) {
+          fprintf(stderr, "flags & ADAPT_AFTER_CROP=%d\n",
+                          flags & ADAPT_AFTER_CROP);
+        }
+        break;
       case '?':  /* missing argument or character not in optstring */
         short_usage(PROG,stderr);
         exit (2);
@@ -634,6 +647,8 @@ int main(int argc, char **argv)
         exit(99);
     }
   }
+  if((flags & ABSOLUTE_THRESHOLD) && (flags & DO_ITERATIVE_THRESHOLD))
+    fprintf(stderr, "%s: warning: -T has no effect due to -a\n", PROG);
   if(flags & DEBUG_OUTPUT) {
     fprintf(stderr, "================================================================================\n");
     fprintf(stderr, "VERSION=%s\n", VERSION);
@@ -652,6 +667,7 @@ int main(int argc, char **argv)
     fprintf(stderr, "flags & OMIT_DECIMAL=%d\n", flags & OMIT_DECIMAL);
     fprintf(stderr, "flags & PRINT_SPACES=%d\n", flags & PRINT_SPACES);
     fprintf(stderr, "flags & SPC_USE_AVG_DST=%d\n", flags & SPC_USE_AVG_DST);
+    fprintf(stderr, "flags & ADAPT_AFTER_CROP=%d\n", flags & ADAPT_AFTER_CROP);
     fprintf(stderr, "need_pixels = %d\n", need_pixels);
     fprintf(stderr, "min_segment = %d\n", min_segment);
     fprintf(stderr, "min_char_dims = %dx%d\n",min_char_dims.w,min_char_dims.h);
@@ -722,14 +738,12 @@ int main(int argc, char **argv)
 
   /* get minimum and maximum "value" values */
   if((flags & DEBUG_OUTPUT) || (flags & PRINT_INFO)) {
+    double min, max;
+    get_minmaxval(&image, lt, &min, &max);
     fprintf(stderr, "%.2f <= lum <= %.2f (lum should be in [0,255])\n",
-                    get_minval(&image, 0, 0, -1, -1, lt),
-                    get_maxval(&image, 0, 0, -1, -1, lt));
+                    min, max);
   }
 
-  /* adapt threshold to image */
-  thresh = adapt_threshold(&image, thresh, lt, 0, 0, -1, -1, flags);
-
   /* process commands */
   if(flags & VERBOSE) /* then print found commands */ {
     if(optind >= argc-1) {
@@ -759,11 +773,12 @@ int main(int argc, char **argv)
             fprintf(stderr, "\n");
           }
           i++;
-          new_image = dilation(&image, thresh, lt, n);
         } else {
+          n = 1;
           if(flags & VERBOSE) fputs(" processing dilation (1)\n", stderr);
-          new_image = dilation(&image, thresh, lt, 1);
         }
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
+        new_image = dilation(&image, thresh, lt, n);
         imlib_context_set_image(image);
         imlib_free_image();
         image = new_image;
@@ -778,11 +793,12 @@ int main(int argc, char **argv)
             fprintf(stderr, "\n");
           }
           i++;
-          new_image = erosion(&image, thresh, lt, n);
         } else {
+          n = 1;
           if(flags & VERBOSE) fputs(" processing erosion (1)\n", stderr);
-          new_image = erosion(&image, thresh, lt, 1);
         }
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
+        new_image = erosion(&image, thresh, lt, n);
         imlib_context_set_image(image);
         imlib_free_image();
         image = new_image;
@@ -797,11 +813,12 @@ int main(int argc, char **argv)
             fprintf(stderr, "\n");
           }
           i++;
-          new_image = opening(&image, thresh, lt, n);
         } else {
+          n = 1;
           if(flags & VERBOSE) fputs(" processing opening (1)\n", stderr);
-          new_image = opening(&image, thresh, lt, 1);
         }
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
+        new_image = opening(&image, thresh, lt, n);
         imlib_context_set_image(image);
         imlib_free_image();
         image = new_image;
@@ -816,22 +833,25 @@ int main(int argc, char **argv)
             fprintf(stderr, "\n");
           }
           i++;
-          new_image = closing(&image, thresh, lt, n);
         } else {
+          n = 1;
           if(flags & VERBOSE) fputs(" processing closing (1)\n", stderr);
-          new_image = closing(&image, thresh, lt, 1);
         }
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
+        new_image = closing(&image, thresh, lt, n);
         imlib_context_set_image(image);
         imlib_free_image();
         image = new_image;
       } else if(strcasecmp("remove_isolated",argv[i]) == 0) {
         if(flags & VERBOSE) fputs(" processing remove_isolated\n", stderr);
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
         new_image = remove_isolated(&image, thresh, lt);
         imlib_context_set_image(image);
         imlib_free_image();
         image = new_image;
       } else if(strcasecmp("make_mono",argv[i]) == 0) {
         if(flags & VERBOSE) fputs(" processing make_mono\n", stderr);
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
         new_image = make_mono(&image, thresh, lt);
         imlib_context_set_image(image);
         imlib_free_image();
@@ -846,13 +866,14 @@ int main(int argc, char **argv)
             }
             fprintf(stderr, "\n");
           }
-          new_image = white_border(&image, bdwidth);
           i++;
         } else {
+          bdwidth = 1;
           if(flags & VERBOSE)
             fputs(" processing white_border (1)\n", stderr);
-          new_image = white_border(&image, 1);
         }
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
+        new_image = white_border(&image, bdwidth);
         imlib_context_set_image(image);
         imlib_free_image();
         image = new_image;
@@ -866,6 +887,7 @@ int main(int argc, char **argv)
         }
         if(i+1<argc-1) {
           offset = (int) atoi(argv[++i]); /* sideeffect: increment i */
+          thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
           new_image = shear(&image, offset);
           imlib_context_set_image(image);
           imlib_free_image();
@@ -885,6 +907,7 @@ int main(int argc, char **argv)
         }
         if(i+1<argc-1) {
           mask = (int) atoi(argv[++i]); /* sideeffect: increment i */
+          thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
           new_image = set_pixels_filter(&image, thresh, lt, mask);
           imlib_context_set_image(image);
           imlib_free_image();
@@ -905,6 +928,7 @@ int main(int argc, char **argv)
         }
         if(i+1<argc-1) {
           mask = (int) atoi(argv[++i]); /* sideeffect: increment i */
+          thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
           new_image = keep_pixels_filter(&image, thresh, lt, mask);
           imlib_context_set_image(image);
           imlib_free_image();
@@ -927,6 +951,7 @@ int main(int argc, char **argv)
             fprintf(stderr, "\n");
           }
           i+=2; /* skip the arguments to dynamic_threshold */
+          thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
           new_image = dynamic_threshold(&image, thresh, lt, ww, wh);
           imlib_context_set_image(image);
           imlib_free_image();
@@ -938,30 +963,35 @@ int main(int argc, char **argv)
         }
       } else if(strcasecmp("rgb_threshold",argv[i]) == 0) {
         if(flags & VERBOSE) fputs(" processing rgb_threshold\n", stderr);
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
         new_image = make_mono(&image, thresh, MINIMUM);
         imlib_context_set_image(image);
         imlib_free_image();
         image = new_image;
       } else if(strcasecmp("r_threshold",argv[i]) == 0) {
         if(flags & VERBOSE) fputs(" processing r_threshold\n", stderr);
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
         new_image = make_mono(&image, thresh, RED);
         imlib_context_set_image(image);
         imlib_free_image();
         image = new_image;
       } else if(strcasecmp("g_threshold",argv[i]) == 0) {
         if(flags & VERBOSE) fputs(" processing g_threshold\n", stderr);
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
         new_image = make_mono(&image, thresh, GREEN);
         imlib_context_set_image(image);
         imlib_free_image();
         image = new_image;
       } else if(strcasecmp("b_threshold",argv[i]) == 0) {
         if(flags & VERBOSE) fputs(" processing b_threshold\n", stderr);
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
         new_image = make_mono(&image, thresh, BLUE);
         imlib_context_set_image(image);
         imlib_free_image();
         image = new_image;
       } else if(strcasecmp("invert",argv[i]) == 0) {
         if(flags & VERBOSE) fputs(" processing invert\n", stderr);
+        thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
         new_image = invert(&image, thresh, lt);
         imlib_context_set_image(image);
         imlib_free_image();
@@ -984,8 +1014,7 @@ int main(int argc, char **argv)
               fprintf(stderr, " adjusting T1=%.2f and T2=%.2f to image\n",
                               t1, t2);
             }
-            min = get_minval(&image, 0, 0, -1, -1, lt);
-            max = get_maxval(&image, 0, 0, -1, -1, lt);
+            get_minmaxval(&image, lt, &min, &max);
             t1 = min + t1/100.0 * (max - min);
             t2 = min + t2/100.0 * (max - min);
             if(flags & VERBOSE) {
@@ -993,6 +1022,7 @@ int main(int argc, char **argv)
             }
           }
           i+=2; /* skip the arguments to gray_stretch */
+          thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
           new_image = gray_stretch(&image, t1, t2, lt);
           imlib_context_set_image(image);
           imlib_free_image();
@@ -1010,7 +1040,7 @@ int main(int argc, char **argv)
         image = new_image;
       } else if(strcasecmp("crop",argv[i]) == 0) {
         if(i+4<argc-1) {
-          int x, y, cw, ch; /* lw = crop width, lh = crop height */
+          int x, y, cw, ch; /* cw = crop width, ch = crop height */
           x = atoi(argv[i+1]);
           y = atoi(argv[i+2]);
           cw = atoi(argv[i+3]);
@@ -1026,7 +1056,8 @@ int main(int argc, char **argv)
             fprintf(stderr, "\n");
           }
           i += 4; /* skip the arguments to crop */
-          imlib_context_set_image(image);
+          if(!(flags & ADAPT_AFTER_CROP))
+            thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
           new_image = crop(&image, x, y, cw, ch);
           imlib_context_set_image(image);
           imlib_free_image();
@@ -1041,13 +1072,13 @@ int main(int argc, char **argv)
           }
           /* get minimum and maximum "value" values in cropped image */
           if((flags&DEBUG_OUTPUT) || (flags&PRINT_INFO) || (flags&VERBOSE)) {
+            double min, max;
+            get_minmaxval(&image, lt, &min, &max);
             fprintf(stderr, "  %.2f <= lum <= %.2f in cropped image"
-                            " (lum should be in [0,255])\n",
-                            get_minval(&image, 0, 0, -1, -1, lt),
-                            get_maxval(&image, 0, 0, -1, -1, lt));
+                            " (lum should be in [0,255])\n", min, max);
           }
           /* adapt threshold to cropped image */
-          thresh = adapt_threshold(&image, thresh, lt, 0, 0, -1, -1, flags);
+          thresh = adapt_threshold(&image, thresh, lt, flags, UPDATE);
         } else {
           fprintf(stderr, "%s: error: crop command needs 4 arguments\n", PROG);
           exit(99);
@@ -1062,6 +1093,7 @@ int main(int argc, char **argv)
             fprintf(stderr, "\n");
           }
           theta = atof(argv[++i]); /* sideeffect: increment i */
+          thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
           new_image = rotate(&image, theta);
           imlib_context_set_image(image);
           imlib_free_image();
@@ -1111,6 +1143,9 @@ int main(int argc, char **argv)
   /* exit if only image processing shall be done */
   if(flags & PROCESS_ONLY) exit(3);
 
+  /* adapt threshold to image (unless this is already done) */
+  thresh = adapt_threshold(&image, thresh, lt, flags, INITIAL);
+
   if(flags & USE_DEBUG_IMAGE) {
     /* copy processed image to debug image */
     debug_image = make_mono(&image, thresh, lt);
@@ -1607,14 +1642,14 @@ int main(int argc, char **argv)
       /* check horizontal segments (vertical scan, x == middle) */
       d_color.R = d_color.A = 255;
       d_color.G = d_color.B = 0;
-      found_pixels = scanline(&image, &debug_image, middle, digits[d].y1,
+      found_pixels = scanline(&debug_image, middle, digits[d].y1,
                               d_height/3, VERTICAL, d_color, thresh, lt, flags);
       if(found_pixels >= need_pixels) {
         digits[d].digit |= HORIZ_UP; /* add upper segment */
       }
       d_color.G = d_color.A = 255;
       d_color.R = d_color.B = 0;
-      found_pixels = scanline(&image, &debug_image, middle,
+      found_pixels = scanline(&debug_image, middle,
                               digits[d].y1 + d_height/3, d_height/3, VERTICAL,
                               d_color, thresh, lt, flags);
       if(found_pixels >= need_pixels) {
@@ -1622,7 +1657,7 @@ int main(int argc, char **argv)
       }
       d_color.B = d_color.A = 255;
       d_color.R = d_color.G = 0;
-      found_pixels = scanline(&image, &debug_image, middle,
+      found_pixels = scanline(&debug_image, middle,
                               digits[d].y1 + 2*d_height/3, d_height/3, VERTICAL,
                               d_color, thresh, lt, flags);
       if(found_pixels >= need_pixels) {
@@ -1631,7 +1666,7 @@ int main(int argc, char **argv)
       /* check upper vertical segments (horizontal scan, y == quarter) */
       d_color.R = d_color.A = 255;
       d_color.G = d_color.B = 0;
-      found_pixels = scanline(&image, &debug_image, digits[d].x1, quarter,
+      found_pixels = scanline(&debug_image, digits[d].x1, quarter,
                               (digits[d].x2 - digits[d].x1) / 2, HORIZONTAL,
                               d_color, thresh, lt, flags);
       if (found_pixels >= need_pixels) {
@@ -1639,7 +1674,7 @@ int main(int argc, char **argv)
       }
       d_color.G = d_color.A = 255;
       d_color.R = d_color.B = 0;
-      found_pixels = scanline(&image, &debug_image,
+      found_pixels = scanline(&debug_image,
                               (digits[d].x1 + digits[d].x2) / 2 + 1,
                               quarter, (digits[d].x2 - digits[d].x1) / 2 - 1,
                               HORIZONTAL, d_color, thresh, lt, flags);
@@ -1649,7 +1684,7 @@ int main(int argc, char **argv)
       /* check lower vertical segments (horizontal scan, y == three_quarters) */
       d_color.R = d_color.A = 255;
       d_color.G = d_color.B = 0;
-      found_pixels = scanline(&image, &debug_image, digits[d].x1,
+      found_pixels = scanline(&debug_image, digits[d].x1,
                               three_quarters, (digits[d].x2 - digits[d].x1) / 2,
                               HORIZONTAL, d_color, thresh, lt, flags);
       if (found_pixels >= need_pixels) {
@@ -1657,7 +1692,7 @@ int main(int argc, char **argv)
       }
       d_color.G = d_color.A = 255;
       d_color.R = d_color.B = 0;
-      found_pixels = scanline(&image, &debug_image,
+      found_pixels = scanline(&debug_image,
                               (digits[d].x1 + digits[d].x2) / 2 + 1,
                               three_quarters, (digits[d].x2-digits[d].x1)/2 - 1,
                               HORIZONTAL, d_color, thresh, lt, flags);
